//Copyright (c) 1999, Art Gittleman
//This example is provided WITHOUT ANY WARRANTY either expressed or implied.

/* Uses synchronized methods to correct the 
 * error in TallyWrong
 */

public class TallyRight {

  class Buffer {          
    int number = 0;        // the number to increase or decrease
    int previous = 0;
    int total = 0;         // total number of operations performed
    int errors = 0;        // number of errors

    public synchronized void increment() {
      number++;
      report(number);
    }

    public synchronized void decrement() {
      number--;
      report(number);
    }

    public void report(int n) {
      total++;
      if (n == previous)
        System.err.println(++errors + "\t" +total);
      System.err.print(" "+n);
      if (total % 100==0)
	System.err.println("\nTotal Operations just passed "+total);
      previous = n;
    }
  }  // Buffer

  class Plus extends Thread {
     Buffer buf;

     Plus(Buffer b) { 
       buf=b;
     }			

     public void run() {
       while (true)
          buf.increment();
       }
  } // Plus

  class Minus extends Thread {
     Buffer buf;

     Minus(Buffer b) { 
       buf=b; 
     }

     public void run() { 
       while (true)
         buf.decrement();
     }
  }  // Minus

  public static void main(String[] argv) {
    TallyRight tr = new TallyRight();
    Buffer b = tr.new Buffer();
    Plus p = tr.new Plus(b);
    Minus m = tr.new Minus(b);
    p.start();
    m.start();
  }
}

